/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.model;

import java.io.*;

/**
 * Class that reads LITTLE endian binary files
 *
 * @author Gerben Feenstra & Merijn Zengers
 */
public class BinaryReader implements Closeable {
    
    private InputStream inputStream = null;

    public BinaryReader(InputStream inputStream) {
        this.inputStream = inputStream;
    }

    /**
     * Method that reads from the binairyFile given to the constructor Reads the
     * number of bytes specified in XmlField.BinaryDataType Will reverse the
     * bytes to get Big endian
     *
     * @param field The field to read
     * @return value that has been red
     */
    public long read(XmlField field) throws IOException {
        assert inputStream != null;
        BinaryDataType datatype = field.getDataType();
        byte[] bytes = new byte[datatype.getSize()];
        inputStream.read(bytes);
        return convert(bytes, datatype);
    }
        
    public static long convert(byte[] bytes, BinaryDataType datatype) throws IOException {
        long value = 0;
        ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
        DataInputStream dataInputStream = new DataInputStream(bais);
        if (dataInputStream.available() == datatype.getSize()) {
            switch (datatype) {
                case BIT8_SIGNED:
                    value = dataInputStream.readByte();
                    break;
                case BIT8_UNSIGNED:
                    value = dataInputStream.readUnsignedByte();
                    break;
                case BIT16_SIGNED:
                    Short signed16temp = dataInputStream.readShort();
                    value = (long) Short.reverseBytes(signed16temp);
                    break;
                case BIT16_UNSIGNED:
                    Short unsigned16temp = dataInputStream.readShort();
                    value = (long) Short.reverseBytes(unsigned16temp);
                    if (value < 0) {
                        value += 65_536; // 2^16
                    }
                    break;
                case BIT32_SIGNED:
                    Integer signed32temp = dataInputStream.readInt();
                    value = (long) Integer.reverseBytes(signed32temp);
                    break;
                case BIT32_UNSIGNED:
                    Integer unsigned32temp = dataInputStream.readInt();
                    value = (long) Integer.reverseBytes(unsigned32temp);
                    if (value < 0) {
                        value += 4_294_967_296L; // 2^32
                    }
                    break;
            }
        }
        return value;
    }

    @Override
    public void close() throws IOException {
        inputStream.close();
        inputStream = null;
    }

    /**
     *
     * @return True if the inputstream != null and has still bits available.
     * @throws IOException
     */
    public boolean available() throws IOException {
        return inputStream != null && inputStream.available() > 0;
    }
}
