/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.model;

import java.util.Arrays;
import nl.cloudfarming.client.util.UOM;

/**
 * Enum which describes all the DDI as formulated in iso 11783-11
 *
 * latest version is available at http://dictionary.isobus.net/isobus/
 *
 * @author johan
 */
public enum DDI {

    /**
     * This DDE is used to specify which version of the Data Dictionary is being
     * used. Range 0 - 2147483647
     */
    DDI_0("This DDE is used to specify which version of the Data Dictionary is being used.", UOM.NOT_SPECIFIED, new byte[]{0, 0}, 1),
    /**
     * Setpoint Application Rate specified as volume per area Range 0 -
     * 2147483647
     */
    DDI_1("Setpoint Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 1}, 0.01),
    /**
     * Actual Application Rate specified as volume per area Range 0 - 2147483647
     */
    DDI_2("Actual Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 2}, 0.01),
    /**
     * Default Application Rate specified as volume per area use when missing
     * Position data or outside any Treatment Zone Range 0 - 2147483647
     */
    DDI_3("Default Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 3}, 0.01),
    /**
     * Minimum Application Rate specified as volume per area supplied by device
     * as physical minimum 0 - 2147483647
     */
    DDI_4("Minimum Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 4}, 0.01),
    /**
     * Maximum Application Rate specified as volume per area supplied by device
     * as physical maximum Range 0 - 2147483647
     */
    DDI_5("Maximum Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 5}, 0.01),
    /**
     * Setpoint Application Rate specified as mass per area Range 0 - 2147483647
     */
    DDI_6("Setpoint Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 6}, 1),
    /**
     * Actual Application Rate specified as mass per area Range 0 - 2147483647
     */
    DDI_7("Actual Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 7}, 1),
    /**
     * Default Application Rate specified as mass per area. Use when missing
     * Position data or outside any Treatment Zone. Range 0 - 2147483647
     */
    DDI_8("Default Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 8}, 1),
    /**
     * Minimum Application Rate specified as mass per area supplied by device as
     * physical minimum. Range 0 - 2147483647
     */
    DDI_9("Minimum Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 9}, 1),
    /**
     * Maximum Application Rate specified as mass per area supplied by device as
     * physical maximum. Range 0 - 2147483647
     */
    DDI_10("Maximum Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 10}, 1),
    /**
     * Setpoint Application Rate specified as count per area. Range 0 -
     * 2147483647
     */
    DDI_11("Setpoint Count Per Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 11}, 0.001),
    /**
     * Actual Application Rate specified as count per area. Range 0 - 2147483647
     */
    DDI_12("Actual Count Per Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 12}, 0.001),
    /**
     * Default Count Per Area Application Rate Default Application Rate
     * specified as count per area. Comment use when missing Position data or
     * outside any Treatment Zone Range 0 - 2147483647
     */
    DDI_13("Default Count Per Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 13}, 0.001),
    /**
     * Minimum Application Rate specified as count per area. Comment: supplied by
     * device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_14("Minimum CountPer Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 14}, 0.001),
    /**
     * Maximum Application Rate specified as count per area. 
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_15("Maximum Count Per Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 15}, 0.001),
    /**
     * Setpoint Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Range 0 - 2147483647
     */
    DDI_16("Setpoint Spacing Application Rate", UOM.MM, new byte[]{0, 16}, 1),
    /**
     * Actual Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Range 0 - 2147483647
     */
    DDI_17("Actual Spacing Application Rate", UOM.MM, new byte[]{0, 17}, 1),
    /**
     * Default Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_18("Default Spacing Application Rate", UOM.MM, new byte[]{0, 18}, 1),
    /**
     * Minimum Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_19("Minimum Spacing Application Rate", UOM.MM, new byte[]{0, 19}, 1),
    /**
     * Maximum Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_20("Maximum Spacing Application Rate", UOM.MM, new byte[]{0, 20}, 1),
    /**
     * Setpoint Application Rate specified as volume per volume.
     * Range 0 - 2147483647
     */
    DDI_21("Setpoint Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 21}, 1),
    /**
     * Actual Application Rate specified as volume per volume.
     * Range 0 - 2147483647
     */
    DDI_22("Actual Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 22}, 1),
    /**
     * Default Application Rate specified as volume per volume.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_23("Default Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 23}, 1),
    /**
     * Minimum Application Rate specified as volume per volume.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_24("Minimum Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 24}, 1),
    /**
     * Maximum Application Rate specified as volume per volume.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_25("Maximum Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 25}, 1),
    /**
     * Setpoint Application Rate specified as mass per mass.
     * Range 0 - 2147483647
     */
    DDI_26("Setpoint Mass Per Mass Application Rate", UOM.MG_PER_KG, new byte[]{0, 26}, 1),
    /**
     * Actual Application Rate specified as mass per mass.
     * Range 0 - 2147483647
     */
    DDI_27("Actual Mass Per Mass Application Rate", UOM.MG_PER_KG, new byte[]{0, 27}, 1),
    /**
     * Default Application Rate specified as mass per mass.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_28("Default Mass Per MassA pplication Rate", UOM.MG_PER_KG, new byte[]{0, 28}, 1),
    /**
     * Minimum Application Rate specified as mass per mass.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_29("Minimum Mass Per Mass Application Rate", UOM.MG_PER_KG, new byte[]{0, 29}, 1),
    /**
     * Maximum Application Rate specified as mass per mass.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_30("MaximumMass Per Mass Application Rate", UOM.MG_PER_KG, new byte[]{0, 30}, 1),
    /**
     * Setpoint Application Rate specified as volume per mass.
     * Range 0 - 2147483647
     */
    DDI_31("Setpoint Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0, 31}, 1),
    /**
     * Actual Application Rate specified as volume per mass.
     * Range 0 - 2147483647
     */
    DDI_32("Actual Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0, 32}, 1),
    /**
     * Default Application Rate specified as volume per mass.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_33("Default Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0,33}, 1),
    /**
     * Minimum Application Rate specified as volume per mass.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_34("Minimum Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0,34}, 1),
    /**
     * Maximum Application Rate specified as volume per mass.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_35("Maximum Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0,35}, 1),
    /**
     * Setpoint Application Rate specified as volume per time.
     * Range 0 - 2147483647
     */
    DDI_36("Setpoint Application Rate specified as volume per time", UOM.MM3_PER_SECOND, new byte[]{0,36}, 1),
    /**
     * Actual Application Rate specified as volume per time.
     * Range 0 - 2147483647
     */
    DDI_37("Actual Volume Per Time Application Rate", UOM.MM3_PER_SECOND, new byte[]{0,37}, 1),
    /**
     * Default Application Rate specified as volume per time.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_38("Default Volume Per Time Application Rate", UOM.MM3_PER_SECOND, new byte[]{0,38}, 1),
    /**
     * Minimum Application Rate specified as volume per time.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_39("Minimum Volume Per Time Application Rate", UOM.MM3_PER_SECOND, new byte[]{0,39}, 1),
    /**
     * Maximum Application Rate specified as volume per time.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_40("Maximum Volume Per Time Application Rate", UOM.MM3_PER_SECOND, new byte[]{0,40}, 1),
    /**
     * Setpoint Application Rate specified as mass per time.
     * Range 0 - 2147483647
     */
    DDI_41("Setpoint Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,41}, 1),
    /**
     * Actual Application Rate specified as mass per time.
     * Range 0 - 2147483647
     */
    DDI_42("Actual Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,42}, 1),
    /**
     * Default Application Rate specified as mass per time.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_43("Default Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,43}, 1),
    /**
     * Minimum Application Rate specified as mass per time.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_44("Minimum Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,44}, 1),
    /**
     * Maximum Application Rate specified as mass per time.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_45("Maximum Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,45}, 1),
    /**
     * Setpoint Application Rate specified as count per time.
     * Range 0 - 2147483647
     */
    DDI_46("Setpoint Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,46}, 0.001),
    /**
     * Actual Application Rate specified as count per time.
     * Range 0 - 2147483647
     */
    DDI_47("Actual Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,47}, 0.001),
    /**
     * Default Application Rate specified as count per time.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_48("Default Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,48}, 0.001),
    /**
     * Minimum Application Rate specified as count per time.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_49("Minimum Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,49}, 0.001),
    /**
     * Maximum Application Rate specified as count per time.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_50("Maximum Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,50}, 0.001),
    /**
     * Setpoint Tillage Depth of Device Element below soil surface, value
     * increases with depth.
     * Range 0 - 2147483647
     */
    DDI_51("Setpoint Tillage Depth", UOM.MM, new byte[]{0,51}, 1),
    /**
     * Actual Tillage Depth of Device Element below soil surface, value
     * increases with depth.</br>
     * Range 0 - 2147483647
     */
    DDI_52("Actual Tillage Depth", UOM.MM, new byte[]{0,52}, 1),
    /**
     * Default Tillage Depth of Device Element below soil surface, value
     * increases with depth.</br>
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_53("Default Tillage Depth", UOM.MM, new byte[]{0,53}, 1),
    /**
     * Minimum Tillage Depth of Device Element below soil surface, value
     * increases with depth.</br>
     * Comment: supplied by device as physical minimum
     * Range 0 - 2147483647
     */
    DDI_54("Minimum Tillage Depth", UOM.MM, new byte[]{0,54}, 1),
    /**
     * Maximum Tillage Depth of Device Element below soil surface, value
     * increases with depth.</br>
     * Comment supplied by device as physical maximum
     * Range 0 - 2147483647
     */
    DDI_55("Maximum Tillage Depth", UOM.MM, new byte[]{0,55}, 1),
    /**
     * Setpoint Seeding Depth of Device Element below soil surface, value
     * increases with depth
     * Range 0 - 2147483647
     */
    DDI_56("Setpoint Seeding Depth", UOM.MM, new byte[]{0,56}, 1),
    /**
     * Definition Actual Seeding Depth of Device Element below soil surface, value
     * increases with depth
     * Range 0 - 2147483647
     */
    DDI_57("Actual Seeding Depth", UOM.MM, new byte[]{0,57}, 1),
    /**
     * Default Seeding Depth of Device Element below soil surface, value
     * increases with depth
     * Comment use when missing Position data or outside any Treatment Zone
     * Range 0 - 2147483647
     */
    DDI_58("Default Seeding Depth", UOM.MM, new byte[]{0,58}, 1),
    /**
     * Minimum Seeding Depth of Device Element below soil surface, value
     * increases with depth
     * Comment supplied by device as physical minimum
     * Range 0 - 2147483647
     */
    DDI_59("Minimum Seeding Depth", UOM.MM, new byte[]{0,59}, 1),
    /**
     * Maximum Seeding Depth of Device Element below soil surface, value
     * increases with depth
     * Comment supplied by device as physical maximum
     * Range 0 - 2147483647
     */
    DDI_60("Maximum Seeding Depth", UOM.MM, new byte[]{0,60}, 1),
    /**
     * Setpoint Working Height of Device Element above crop or soil
     * Range 0 - 2147483647
     */
    DDI_61("Setpoint Working Height", UOM.MM, new byte[]{0,61}, 1),
    /**
     * Actual Working Height of Device Element above crop or soil
     * Comment This is the height above the effective control surface. For sprayers
     * this is the height above the crop canapé and for fertilizer spreaders,
     * harvesters, etc it is the height above the ground.
     * Range 0 - 2147483647
     */
    DDI_62("Actual Working Height", UOM.MM, new byte[]{0,62}, 1),
    /**
     * Default Working Height of Device Element above crop or soil
     * Comment use when missing Position data or outside any Treatment Zone
     * Range 0 - 2147483647
     */
    DDI_63("Default Working Height", UOM.MM, new byte[]{0,63}, 1),
    /**
     * Minimum Working Height of Device Element above crop or soil
     * Comment supplied by device as physical minimum
     * Range 0 - 2147483647
     */
    DDI_64("Minimum Working Height", UOM.MM, new byte[]{0,64}, 1),
    /**
     * Maximum Working Height of Device Element above crop or soil
     * Comment supplied by device as physical maximum
     * Range 0 - 2147483647
     */
    DDI_65("Maximum Working Height", UOM.MM, new byte[]{0,65}, 1),
    /**
     * Setpoint Working Width of Device Element
     * Range 0 - 2147483647
     */
    DDI_66("Setpoint Working Width", UOM.MM, new byte[]{0,66}, 1),
    /**
     * Actual Working Width of Device Element.
     * Comment: This is the effective / active working width during operation.
     * Range 0 - 2147483647
     */
    DDI_67("Actual Working Width", UOM.MM, new byte[]{0, 67}, 1),
    /**
     * Default Working Width of Device Element
     * Range 0 - 2147483647
     */
    DDI_68("Default Working Width", UOM.MM, new byte[]{0, 68}, 1),
    /**
     * Minimum Working Width of Device Element
     * Range 0 - 2147483647
     */
    DDI_69("Minimum Working Width", UOM.MM, new byte[]{0, 69}, 1),
    /**
     * Maximum Working Width of Device Element
     * Range 0 - 2147483647
     */
    DDI_70("Maximum Working Width", UOM.MM, new byte[]{0, 70}, 1),
    /**
     * Setpoint Device Element Content specified as volume
     * Range 0 - 2147483647
     */
    DDI_71("Setpoint Volume Content", UOM.ML, new byte[]{0, 71}, 1),
    /**
     * Actual Device Element Content specified as volume
     * Range 0 - 2147483647
     */
    DDI_72("Actual Volume Content", UOM.ML, new byte[]{0, 72}, 1),
    /**
     * Maximum Device Element Content specified as volume
     * Comment is a minimum needed as well ??
     * Range 0 - 2147483647
     */
    DDI_73("Maximum Volume Content", UOM.ML, new byte[]{0, 73}, 1),
    /**
     * Setpoint Machine Element Content specified as mass
     * Range 0 - 2147483647
     */
    DDI_74("Setpoint Mass Content", UOM.GRAM, new byte[]{0, 74}, 1),
    /**
     * Actual Device Element Content specified as mass
     * Comment If the device is equipped with a weighing system which provides the
     * possibility to tare the current load it is possible that the value has a
     * negative sign in case of an unload operation.
     * Range -2147483648 - 2147483647
     */
    DDI_75("Actual Mass Content", UOM.GRAM, new byte[]{0, 75}, 1),
    /**
     * Maximum Device Element Content specified as mass
     * Comment is a minimum needed as well ??
     * Range 0 - 2147483647
     */
    DDI_76("Maximum Mass Content", UOM.GRAM, new byte[]{0, 76}, 1),
    /**
     * Setpoint Device Element Content specified as count
     * Range 0 - 2147483647
     */
    DDI_77("Setpoint Device Element Content specified as count", UOM.COUNT, new byte[]{0, 77}, 1),
    /**
     * <p>Actual Device Element Content specified as count</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_78("Actual Count Content", UOM.COUNT, new byte[]{0, 78}, 1),
    /**
     * <p>Maximum Device Element Content specified as count</p>
     * <p>Comment: is a minimum needed as well ??</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_79("Maximum Count Content", UOM.COUNT, new byte[]{0, 79}, 1),
    /**
     * <p>Accumulated Application specified as volume</p>
     * <p>Comment: is a counter of a device element</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_80("Application Total Volume", UOM.LITERS, new byte[]{0, 80}, 1),
    /**
     * <p>Accumulated Application specified as mass.</p>
     * <p>Comment: is a counter of a device element.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_81("Application Total Mass", UOM.KG, new byte[]{0, 81}, 1),
    /**
     * <p>Accumulated Application specified as count</p>
     * <p>Comment: is a counter of a device element</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_82("Application Total Count", UOM.COUNT, new byte[]{0, 82}, 1),
    /**
     * <p>Yield as volume per area</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_83("Volume Per Area Yield", UOM.ML_PER_M2, new byte[]{0, 83}, 1),
    /**
     * <p>
     * Yield as mass per area, not corrected for the reference moisture
     * percentage DDI 184.
     * </p>
     * <p>
     * Comment:
     * This Mass per Area yield is the mass that includes the actual
     * percentage moisture (DDI 99) if this is measured on e.g. harvesting
     * equipment. This comment is added to clarify and differentiate this DDI
     * from the Dry Mass Per Area Yield (DDI 181).
     * </p>
     * <p>
     * Range 0 - 2147483647
     * </p>
     */
    DDI_84("Mass Per Area Yield", UOM.MG_PER_M2, new byte[]{0, 84}, 1),
    /**
     * <p>Definition Yield as count per area</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_85("Count Per Area Yield", UOM.AMOUNT_PER_M2, new byte[]{0, 85}, 1),
    /**
     * <p>Yield as volume per time</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_86("Volume Per Time Yield", UOM.ML_PER_SEC, new byte[]{0, 86}, 1),
    /**
     * <p>Yield as mass per time, not corrected for the reference moisture
     * percentage DDI 184.</p>
     * <p>Comment: This Mass per Time yield is the mass that includes the actual
     * percentage moisture (DDI 99) if this is measured on e.g. harvesting
     * equipment. This comment is added to clarify and differentiate this DDI
     * from the Dry Mass Per Time Yield (DDI 182).</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_87("Mass Per Time Yield", UOM.MG_PER_SECOND, new byte[]{0, 87}, 1),
    /**
     * <p>Yield as count per time</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_88("Count Per Time Yield", UOM.AMOUNT_PER_SECOND, new byte[]{0, 88}, 1),
    /**
     * <p>Accumulated Yield specified as volume</p>
     * <p>Comment: is a counter of a machine element</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_89("Yield Total Volume", UOM.LITERS, new byte[]{0, 89}, 1),
    /**
     * <p>Accumulated Yield specified as mass, not corrected for the reference
     * moisture percentage DDI 184.</p>
     * <p>Comment: This Yield Total Mass is the mass that includes the actual percentage
     * moisture (DDI 99) if this is measured on e.g. harvesting equipment.
     * This comment is added to clarify and differentiate this DDI from the
     * Yield Total Dry Mass (DDI 183).</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_90("Yield Total Mass", UOM.KG, new byte[]{0, 90}, 1),
    /**
     * <p>Accumulated Yield specified as count</p>
     * <p>Comment: is a counter of a machine element</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_91("Yield Total Count", UOM.COUNT, new byte[]{0, 91}, 1),
    /**
     * <p>Crop yield loss as volume per area</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_92("Volume Per Area Crop Loss", UOM.ML_PER_M2, new byte[]{0, 92}, 1),
    /**
     * <p>Crop yield loss as mass per area</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_93("Mass Per Area Crop Loss", UOM.ML_PER_M2, new byte[]{0, 93}, 1),
    /**
     * <p>Crop yield loss as count per area</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_94("Count Per Area Crop Loss", UOM.AMOUNT_PER_M2, new byte[]{0, 94}, 0.001),
    /**
     * <p>Crop yield loss as volume per time</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_95("Volume Per Time Crop Loss", UOM.ML_PER_SEC, new byte[]{0, 95}, 1),
    /**
     * <p>Crop yield loss as mass per time</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_96("Mass Per Time Crop Loss", UOM.MG_PER_SECOND, new byte[]{0, 96}, 1),
    /**
     * <p>Crop yield loss as count per time</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_97("Count Per Time Crop Loss", UOM.AMOUNT_PER_SECOND, new byte[]{0, 97}, 0.001),
    /**
     * <p>Crop yield loss</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_98("Percentage Crop Loss", UOM.PARTS_PER_MILLION, new byte[]{0, 98}, 1),
    /**
     * <p>Moisture in crop yield</p>
     * <p>Comment This DDE defines the actual percentage moisture of the crop.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_99("Crop Moisture", UOM.PARTS_PER_MILLION, new byte[]{0, 99}, 1),
    /**
     * <p>Dirt or foreign material in crop yield</p>
     * <p>Comment This DDE defines the contamination in ratio of the yield DDI units.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_100("Crop Contamination", UOM.PARTS_PER_MILLION, new byte[]{0, 100}, 1),
    /**
     * <p>Setpoint Bale Width for square baler or round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_101("Setpoint Bale Width", UOM.MM, new byte[]{0, 101}, 1),
    /**
     * <p>Actual Bale Width for square baler or round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_102("Actual Bale Width", UOM.MM, new byte[]{0, 102}, 1),
    /**
     * <p>Default Bale Width for square baler or round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_103("Default Bale Width", UOM.MM, new byte[]{0, 103}, 1),
    /**
     * <p>Minimum Bale Width for square baler or round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_104("Minimum Bale Width", UOM.MM, new byte[]{0, 104}, 1),
    /**
     * <p>Maximum Bale Width for square baler or round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_105("Maximum Bale Width", UOM.MM, new byte[]{0, 105}, 1),
    /**
     * <p>Setpoint Bale Height is only applicable to square baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_106("Setpoint Bale Height", UOM.MM, new byte[]{0, 106}, 1),
    /**
     * <p>Actual Bale Height is only applicable to square baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_107("ActualBaleHeight", UOM.MM, new byte[]{0, 107}, 1),
    /**
     * <p>Default Bale Height is only applicable to square baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_108("Default Bale Height", UOM.MM, new byte[]{0, 108}, 1),
    /**
     * <p>Minimum Bale Height is only applicable to square baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_109("Minimum Bale Height", UOM.MM, new byte[]{0, 109}, 1),
    /**
     * <p>Maximum Bale Height is only applicable to square baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_110("Maximum Bale Height", UOM.MM, new byte[]{0, 110}, 1),
    /**
     * <p>Setpoint Bale Size as length for a square baler or diameter for a
     * round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_111("Setpoint Bale Size", UOM.MM, new byte[]{0, 111}, 1),
    /**
     * <p>Actual Bale Size as length for a square baler or diameter for a round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_112("Actual Bale Size", UOM.MM, new byte[]{0, 112}, 1),
    /**
     * <p>Default Bale Size as length for a square baler or diameter for a round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_113("Default Bale Size", UOM.MM, new byte[]{0, 113}, 1),
    /**
     * <p>Minimum Bale Size as length for a square baler or diameter for a round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_114("Minimum Bale Size", UOM.MM, new byte[]{0, 114}, 1),
    /**
     * <p>Maximum Bale Size as length for a square baler or diameter for a round baler</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_115("Maximum Bale Size", UOM.MM, new byte[]{0, 115}, 1),
    /**
     * <p>Accumulated Area.</p>
     * <p>Comment: is a counter of a machine element.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_116("Total Area", UOM.M2, new byte[]{0, 116}),
    /**
     * <p>Accumulated Distance in working position.</p>
     * <p>Comment: is a counter of a machine element.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_117("Effective Total Distance", UOM.MM, new byte[]{0, 117}, 1),
    /**
     * <p>Accumulated Distance out of working position.</p>
     * <p>Comment: is a counter of a machine element.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_118("Ineffective Total Distance", UOM.MM, new byte[]{0, 118}, 1),
    /**
     * <p>Definition Accumulated Time in working position.</p>
     * <p>Comment: is a counter of a machine element.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_119("Effective Total Time", UOM.SECOND, new byte[]{0, 119}, 1),
    /**
     * <p>Accumulated Time out of working position.</p>
     * <p>Comment: is a counter of a machine element.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_120("Ineffective Total Time", UOM.SECOND, new byte[]{0, 120}, 1),
    /**
     * <p>Product Density as mass per volume</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_121("Product Density Mass Per Volume", UOM.MG_PER_L, new byte[]{0, 121}, 1),
    /**
     * <p>X direction offset of a DeviceElement relative to a Device.</p>
     * <p>Range -2147483648 - 2147483647</p>
     */
    DDI_134("Device Element Offset X", UOM.MM, new byte[]{0, Byte.parseByte("-122")}, 1),
    /**
     * <p>Y direction offset of a DeviceElement relative to a Device.</p>
     * <p>Range -2147483648 - 2147483647</p>
     */
    DDI_135("Device Element Offset Y", UOM.MM, new byte[]{0, Byte.parseByte("-121")}, 1),
    /**
     * <p>Z direction offset of a DeviceElement relative to a Device.</p>
     * <p>Comment: This DDE was a result of Jan 2005 Task Controller meeting.</p>
     * <p>Range -2147483648 - 2147483647</p>
     */
    DDI_136("Device Element Offset Z", UOM.MM, new byte[]{0, Byte.parseByte("-120")}, 1),
    /**
     * <p>DeviceElement Volume Capacity, dimension of a DeviceElement</p>
     * <p>Comment This DDE was a result of Jan 2005 Task Controller meeting.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_137("Device Volume Capacity", UOM.ML, new byte[]{0, Byte.parseByte("-119")}, 1),
    /**
     * <p>DeviceElement Mass Capacity, dimension of a DeviceElement</p>
     * <p>Comment This DDE was a result of Jan 2005 Task Controller meeting.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_138("Device Mass Capacity", UOM.GRAM, new byte[]{0, Byte.parseByte("-118")}, 1),
    /**
     * <p>DeviceElement Count Capacity, dimension of a DeviceElement</p>
     * <p>Comment This DDE was a result of Jan 2005 Task Controller meeting.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_139("Device Count Capacity", UOM.COUNT, new byte[]{0, Byte.parseByte("-117")}, 1),
    /**
     * <p>Application Rate expressed as percentage</p>
     * <p>Comment This DDE was a result of Jan 2005 Task Controller meeting.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_140("Percentage Application Rate Setpoint", UOM.PARTS_PER_MILLION, new byte[]{0, Byte.parseByte("-116")}, 1),
    /**
     * <p>Actual Work State, 2 bits defined as 00=disabled/off, 01=enabled/on,
     * 10=error, 11=undefined/not installed. See the DDI 289 attachment
     * "ISO11783-11-DDI-289-SetpointWorkState" for implementation guidelines.
     * This DDE has been revised in 2012 to be used as the Actual Work State. A
     * separate Setpoint Work State was added to the data dictionary at that
     * time. Original comment: this DDE was a result of March 2005 TF10 meeting.</p>
     * <p>Range 0-3</p>
     */
    DDI_141("Actual Work State", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-115")}, 1), // AKA on/off
    /**
     * The Setpoint Value Latency Time is the time lapse between the moment of
     * receival of a setpoint value command by the working set and the moment
     * this setpoint value is physically applied on the device. That means if
     * the setpoint value is communicated on the network (CAN bus) but the
     * system needs 2 seconds to adjust the value physically on the desired unit
     * (device element) then the Setpoint Latency Time is 2 seconds. The
     * setpoint time latency value can only be positive.</br> The use of this
     * DDE is to inform the overall system (e.g. Dektop Software, Task
     * Controller) how the system works. The TC shall not shift this information
     * into log files nor shall the device do that when sending actual values.
     *
     */
    DDI_142("Physical Setpoint Time Latency", UOM.MS, new byte[]{0, Byte.parseByte("-114")}, 1),
    /**
     * <p>The Actual Value Latency Time is the time lapse between the
     * moment this actual value is physically applied on the device, and the
     * moment that this actual value is communicated to the Task
     * Controller. That means if the system needs 2 seconds to calculate or
     * measure a value before communicating it on the network, then the
     * Actual Latency Time value is 2 seconds.
     * Depending of the system characteristics the latency time could be
     * negative or positive. In case where the system communicates an actual value before the
     * actual value has been physically applied the latency value should be positive.
     * In case where the system communicates an actual value after the
     * actual value has been physically applied the latency value should be negative.</p>
     * <p>The use of this DDE is to inform the overall system (e.g. Dektop
     * Software, Task Controller) how the system works. The TC shall not
     * shift this information into log files nor shall the device do that when
     * sending actual values.
     * Example for a positive value:
     * A seed flow sensor is placed at the start of the seed tube. At the
     * moment the sensor measures seed flow X, it takes Y seconds for this
     * flow to reach the coulters. The measured value is ahead of the
     * physical value of the unit (coulters). So the latency could be plus 2
     * seconds.
     * Example for a negative value:
     * A flow sensor has a delay in its response to a flow change that
     * means it takes Y seconds to realize the change. At the moment the
     * sensor measures flow X, the flow is already present for Y seconds on
     * the physical unit. The actual value is ahead of measured value. So
     * the latency could be minus 2 seconds.</p>
     * <p>Range -2147483648 - 2147483647</p>
     */
    DDI_143("Physical Actual Value Time Latency", UOM.MS, new byte[]{0, Byte.parseByte("-113")}, 1),
    /**
     * <p>Pivot / Yaw Angle of a DeviceElement</p>
     * <p>This DDE was a result of March 2005 TF10 meeting</p>
     * <p>Range -180000 - 180000</p>
     */
    DDI_144("Yaw Angle", UOM.ANGLE, new byte[]{0, Byte.parseByte("-112")}, 0.001),
    /**
     * <p>Roll Angle of a DeviceElement</p>
     * <p>This DDE was a result of March 2005 TF10 meeting</p>
     * <p>Range -180000 - 180000</p>
     */
    DDI_145("Roll Angle", UOM.ANGLE, new byte[]{0, Byte.parseByte("-111")}, 0.001),
    /**
     * <p>Pitch Angle of a DeviceElement</p>
     * <p>This DDE was a result of March 2005 TF10 meeting</p>
     * <p>Range -180000 - 180000</p>
     */
    DDI_146("Pitch Angle", UOM.ANGLE, new byte[]{0, Byte.parseByte("-110")}, 0.001),
    /**
     * <p>Log Counter, may be used to control data log record generation on a
     * Task Controller</p>
     * <p>This DDE was a result of March 2005 TF10 meeting</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_147("Log Count", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-109")}, 1),
    /**
     * <p>Accumulated Fuel Consumption as Counter</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_148("Total Fuel Consumption", UOM.ML, new byte[]{0, Byte.parseByte("-108")}, 1),
    /**
     * <p>Fuel consumption per time</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_149("Instantaneous Fuel Consumption per Time", UOM.MM3_PER_SECOND, new byte[]{0, Byte.parseByte("-107")}, 1),
    /**
     * <p>Fuel consumption per area</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_150("Instantaneous Fuel Consumption per Area", UOM.MM3_PER_M2, new byte[]{0, Byte.parseByte("-106")}, 1),
    /**
     * <p>Area per time capacity</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_151("Instantaneous Area Per Time Capacity", UOM.MM2_PER_SECOND, new byte[]{0, Byte.parseByte("-105")}, 1),
    /**
     * <p>The Normalized Difference Vegetative Index (NDVI) computed from
     * crop reflectances as the difference between NIR reflectance in the
     * 780 to 880 nm band and red reflectance in the 640 to 680 nm band
     * divided by the sum of the NIR and red reflectance in the same bands.</p>
     * <p>See http://dictionary.isobus.net/isobus/attachments/173/NDVI_definition_summary.pdf</p>
     * <p>Range -1 - 1</p>
     */
    DDI_153("Actual Normalized Difference Vegetative Index (NDVI)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-103")}, 0.001),
    /**
     * <p>Length of device element (dimension along the X-axis)</p>
     * <p>Comment The reference point of the device element shall be located in the
     * center of the device element</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_154("Physical Object Length", UOM.MM, new byte[]{0, Byte.parseByte("-102")}, 1),
     /**
     * <p>Width of device element (dimension along the Y-axis)</p>
     * <p>The reference point of the device element shall be located in the
     * center of the device element</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_155("Physical Object Width", UOM.MM, new byte[]{0, Byte.parseByte("-101")}, 1),
    /**
     * <p>Width of device element (dimension along the Y-axis)</p>
     * <p>The reference point of the device element shall be located in the
     * center of the device element</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_156("Height of device element (dimension along the Z-axis)", UOM.MM, new byte[]{0, Byte.parseByte("-100")}, 1),
    /**
     * <p>Specification of the type of coupler. The value definitions are:
     * 0 = unknown (default),
     * 1 = ISO 6489-3 Tractor drawbar,
     * 2 = ISO 730 Three-point-hitch semi-mounted,
     * 3 = ISO 730 Three-point-hitch mounted,
     * 4 = ISO 6489-1 Hitch-hook,
     * 5 = ISO 6489-2 Clevis coupling 40,
     * 6 = ISO 6489-4 Piton type coupling,
     * 7 = ISO 6489-5 CUNA hitch, or ISO 5692-2 Pivot wagon hitch
     * 8 = ISO 24347 Ball type hitch
     * all other values are reserved for future assignments.</p>
     * <p>Comment: This DDE allows systems to automatically select the connection
     * between devices. For instance, when 2 devices are on the network
     * that declare device elements with the same connector type, the
     * system can connect them accordingly.
     * This DDE shall be used with the Device Element of type "Connector"
     * only.</p>
     * <p>Range 0 - 8</p>
     */
    DDI_157("Connector Type", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-99")}, 1),
    /**
     * <p>Defines and synchronise the actual state of the prescription system.
     * The state is represented by the lowest significant 2 bits in the lowest
     * significant byte of the process data value: Byte 1: bit 0-1: 00 =
     * manual/off, 01 = auto/on, 10 = error indicator, 11 = undefined/not
     * installed. bits 2-7: reserved set to 0. Byte 2-4: reserved set to 0.
     * See attachment for more information 
     * http://dictionary.isobus.net/isobus/attachments/203/PCS Implementation Example.pdf</p>
     * <p>Comment: The prescription control master and its clients need to be
     * synchronized in terms of their general state or activation by the user
     * (System activated/deactivated in individual setups). This DDE serves
     * 2 purposes, one is to synchronize the prescription control state and
     * the other is to enable a TC client to announce the support and initial
     * state of its prescription control capabilities. Synchronisation by the
     * prescription master (TC) is done by setting the prescription state of
     * connected TC clients with a process data set value message with this
     * DDE. It is recommended that TC clients reply their state immediately
     * (within 250 ms) when such a message is received. The property flag
     * "setable" and the trigger method "on change" shall be used with this
     * DDE. The state "manual/off" indicates that the device is in manual
     * state and will ignore all prescription commands. The "auto/on" state
     * indicates that the client accepts the prescription commands as far as
     * its overall process state allows.</p>
     * <p>Range 0 - 3</p>
     */
    DDI_158("Prescription Control State", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-98")}, 1),
    /**
     * <p>Specifies the number of sub-units for a section (e.g. number of
     * nozzles per sprayer section or number of planter row units per
     * metering device).</p>
     * <p>Comment: This DDE is used for objects which have further sub-units per section,
     * which are of interest to the operator but not needed for the Task
     * Controller operation itself. With this information and the overall width
     * of the section the system can calculate for instance the row spacing
     * without having individual objects for each row in the DCD.</p>
     * <p>Range 0 - 2147483647</p>
     */
    DDI_159("Number of Sub-Units per Section", UOM.COUNT, new byte[]{0, Byte.parseByte("-97")}, 1),
    /**
     * Specifies the actual state of section control. The value definitions are:
     * Byte 1 (bitfield) Bit 0-1: 00 = manual/off, 01 = auto/on, 10 = error
     * indicator, 11 = undefined/not installed. Bits 2-7: reserved, set to 0.
     * Bytes 2-4: reserved, set to 0. See attachment for more information</br>
     * In section control systems, the section control master and its clients
     * need to be synchronized in terms of their general state or activation by
     * the user (System activated/deactivated in individual setups). This DDE
     * serves 2 purposes, one is to synchronize the section control state and
     * the other is to enable a TC client to announce the support and initial
     * state of its section control capabilities. Synchronisation by the section
     * control master (TC) is done by setting the section control state of
     * connected TC clients with a process data set value message with this DDE.
     * It is recommended that TC clients reply their state immediately (within
     * 250 ms) when such a message is received. The property flag "setable" and
     * the trigger method "on change" should be used with this DDE. The state
     * "manual/off" indicates that the device is in manual state and will ignore
     * all control commands for section control. The "auto/on" state indicates
     * that the client accepts the section control commands as far as its
     * overall process state allows. Listed below are 4 example Use Cases for
     * this DDE: Use case "Start up operation": 1. During a start up the
     * implement shall set the SCS to ‘manual mode’. Use case "Auto request from
     * TC": 1. The implement receives an ‘auto’ request from TC. 2. The
     * implement shall check whether all setup conditions are fulfilled to allow
     * section control. 3. If this check is ok: The implement may respond with
     * ‘auto mode’ and set its internal SCS client to ‘auto mode’. 4. If this
     * check not ok: The implement shall respond with ‘manual mode’. The
     * internal state is still in manual mode. The TC may inform the operator
     * accordingly. Use case "Manual request from TC": 1. The implement receives
     * a manual request from TC. 2. If the implement is still in ‘auto mode’ the
     * implement shall set its internal SCS client to ‘manual mode’. Use case
     * "Loss of requirements for auto mode": 1. The implement internal setup
     * conditions don’t allow for automatic section control anymore. 2. The
     * implement shall set the internal SCS client to ‘manual mode’. 3. The
     * implement shall send the SCS to inform the TC accordingly. On reception
     * of this ‘manual mode’ the TC/Section Control Master may inform the
     * operator accordingly.
     *
     */
    DDI_160("Section Control State", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-96")}, 1),
    /**
     * Combination of the actual work states of individual sections or units
     * (e.g. nozzles) number 1 to 16 into a single actual work state of their
     * parent device element. The actual condensed work state contains the
     * child element actual work states, in the driving direction from left to
     * right, where the leftmost child element actual work state are the 2
     * lowest significant bits of the Process Data Value. Each child device
     * elements actual work state is represented by 2 bits and defined as:
     * 00 = disabled/off, 01 = enabled/on, 10 = error indicator, 11 =
     * undefined/not installed. In total 16 child device element actual work
     * states can be contained in one actual condensed work state of their
     * parent device element. If less than 16 child device element actual
     * work states are available, then the unused bits shall be set to value
     * 11 (not installed). When the parent device element contains the
     * Actual Condensed Work State DDE, then the device descriptor shall
     * not contain the individual actual work state DDEs of the child device
     * elements.
     * 
     */
    DDI_161("Actual Condensed Work State (1-16)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-95")}),
    // similar to DDI 161, repeat for sections/units/nozzles 17 to 256, 16 nozzles per DDI:
    DDI_162("Actual Condensed Work State (17-32)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-94")}),
    DDI_163("Actual Condensed Work State (33-48)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-93")}),
    DDI_164("Actual Condensed Work State (49-64)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-92")}),
    DDI_165("Actual Condensed Work State (65-80)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-91")}),
    DDI_166("Actual Condensed Work State (81-96)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-90")}),
    DDI_167("Actual Condensed Work State (97-112)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-89")}),
    DDI_168("Actual Condensed Work State (113-128)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-88")}),
    DDI_169("Actual Condensed Work State (129-144)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-87")}),
    DDI_170("Actual Condensed Work State (145-160)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-86")}),
    DDI_171("Actual Condensed Work State (161-176)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-85")}),
    DDI_172("Actual Condensed Work State (177-192)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-84")}),
    DDI_173("Actual Condensed Work State (193-208)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-83")}),
    DDI_174("Actual Condensed Work State (209-224)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-82")}),
    DDI_175("Actual Condensed Work State (225-240)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-81")}),
    DDI_176("Actual Condensed Work State (241-256)", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-80")}),
    /**
     * Actual length of cut for harvested material.
     * e.g. Forage Harvester or Tree Harvester.
     */
    DDI_177("Actual length of cut", UOM.MM, new byte[]{0, Byte.parseByte("-79")}, 0.001),
    /**
     * This DDI is used to enumerate and identify multiple device elements
     * (DET) of the same type within one Device Description object pool.
     * The value of this DDI is independent of the DET number. The
     * combination of device element type and value of Element Type
     * Instance ETI represents a unique object inside the device description
     * object pool and therefore shall exist only once per object pool.
     * Recommendation: The definition of the device elements should be
     * made from left to right direction or from front to back direction. When
     * in a matrix, count left-to-right first, then front-to-back and at last
     * top-to-bottom.
     */
    DDI_178("Element Type Instance", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-78")}),
    /**
     * This DDI is used to define the current cultural practice which is
     * performed by an individual device operation. For instance a
     * planter/seeder could provide a sowing and a fertilizing operation at
     * the same time.
     * <p>The cultural practice value definitions are:</p>
     * <ul>
     * <li>0=Unknown,
     * <li>1=Fertilizing,
     * <li>2=Sowing and Planting,
     * <li>3=Crop Protection,
     * <li>4=Tillage,
     * <li>5=Baling,
     * <li>6=Mowing,
     * <li>7=Wrapping,
     * <li>8=Harvesting,
     * <li>9=Forage Harvesting,
     * <li>10=Transport,
     * <li>11-255=Reserved for future Assignment
     * </ul>
     */
    DDI_179("Actual Cultural Practice", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-77")}),
    /**
     * This DDI is used to specify the distance from the Device Reference
     * Point (DRP) down to the ground surface. The DRP to Ground DDI
     * shall be attached only to the Device Element (DET) with element
     * number zero.
     */
    DDI_180("Device Reference Point (DRP) to Ground distance", UOM.MM, new byte[]{0, Byte.parseByte("-76")}),
    /**
     * Yield as dry mass per area. The definition of dry mass is the mass
     * with a reference moisture specified by DDI 184.
     * <p>
     * The earlier defined DDI 84 is the mass per area that is measured on
     * e.g. harvesting equipment as a mass including a possibly unknown
     * moisture percentage. This DDI 181 is the mass per area yield,
     * corrected to a reference moisture.</p>
     */
    DDI_181("Dry Mass Per Area Yield", UOM.MG_PER_M2, new byte[]{0, Byte.parseByte("-75")}),
    /**
     * Yield as dry mass per time. The definition of dry mass is the mass
     * with a reference moisture specified by DDI 184.
     * <p>
     * The earlier defined DDI 87 is the mass per time that is measured on
     * e.g. harvesting equipment as a mass including a possibly unknown
     * moisture percentage. This DDI 182 is the mass per time yield,
     * corrected to a reference moisture.</p>
     */
    DDI_182("Dry Mass Per Time Yield", UOM.MG_PER_SECOND, new byte[]{0, Byte.parseByte("-74")}),
    /**
     * Accumulated Yield specified as dry mass. The definition of dry mass
     * is the mass with a reference moisture specified by DDI 184.
     * <p>
     * The earlier defined DDI 90 is considered to be the total mass that is
     * measured on e.g. harvesting equipment as a mass including a
     * possibly unknown moisture percentage. This DDI 183 is the yield total
     * mass, corrected to a reference moisture.</p>
     */
    DDI_183("Yield Total Dry Mass", UOM.KG, new byte[]{0, Byte.parseByte("-73")}),
    /**
     * Moisture percentage used for the dry mass DDIs 181, 182 and 183.
     * Example: this definition is similar to the "Standard Payable Moisture"
     * term used by farmers.
     */
    DDI_184(" Reference Moisture For Dry Mass", UOM.PARTS_PER_MILLION, new byte[]{0, Byte.parseByte("-72")}),
    
    //      
    // TODO: add more DDI values...
    //
    
    /**
     * Request Default Process Data. This DDE is the highest ISO assigned
     * entity. The range above this number is reserved for manufacture specific
     * DDE's. Range 0-0
     */
    DDI_57343("Request Default Process Data", UOM.NOT_SPECIFIED, new byte[]{Byte.parseByte("-33"), Byte.parseByte("-128")}, 1);
    /*
     * 57344 - 65534 Proprietary DDI Range: Manufacturer proprietary definitions;
     * create separate enums per manufacturer if we need to process these,
     * e.g. Grimme Harvester uses them.
     * 65535 - Reserved.
     */
    public static final double DEFAULT_RESOLUTION = 1.0;
    private final String description;
    private final byte[] DDEntityId; // length 2
    private final UOM uom;
    private final double resolution;

    private DDI(String description, UOM uom, byte[] DDEntityId, double resolution) {
        this.DDEntityId = DDEntityId;
        this.description = description;
        this.uom = uom;
        this.resolution = resolution;
    }

    private DDI(String description, UOM uom, byte[] DDEntityId) {
        this(description, uom, DDEntityId, DEFAULT_RESOLUTION);
    }

    public String getDescription() {
        return description;
    }

    public byte[] getDDEntityId() {
        return DDEntityId;
    }

    public UOM getUom() {
        return uom;
    }

    public double getResolution() {
        return resolution;
    }

    public int toExportValue(double value) {
        return (int) (value / resolution);
    }

    public Double toImportValue(int value) {
        return (value * resolution);
    }

    @Override
    public String toString() {
        String result = description;
        if (uom != null) {
            result += ", " + uom.getLabel();
        }
        return result;
    }

    public static DDI findById(byte[] bytes) {
        for (DDI ddi : DDI.values()) {
            if(Arrays.equals(bytes, ddi.getDDEntityId())) {
                return ddi;
            }
        }
        return null;
    }
    
    public static byte[] hexToBytes(String hexAsString) {
        // via unsigned short:
        int us = Integer.parseInt(hexAsString, 16);
        return new byte[] { (byte)(us>>>8), (byte)us };
    }
}
