/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.model.util;

import com.vividsolutions.jts.geom.Coordinate;
import java.awt.geom.Point2D;
import java.util.*;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DataLogValue;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.DeviceElementTypeEnum;
import nl.cloudfarming.client.isobus.model.DeviceObjectReference;
import nl.cloudfarming.client.isobus.model.DeviceProperty;
import nl.cloudfarming.client.isobus.model.IsobusDevice;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.lib.geotools.GeometryTools;

/**
 * Util that provides functionality to interpret Timelog data
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
public class TimeLogUtil {

    public static final int WORK_STATE_OFF_MANUAL = 0;
    public static final int WORK_STATE_ON_AUTO = 1;
    public static final int WORK_STATE_ERROR = 2;
    public static final int WORK_STATE_NOT_INSTALLED = 3;
    
    public static final DDI[] CONDENSED_WORK_STATE_DDIS = new DDI[]{DDI.DDI_161, DDI.DDI_162, DDI.DDI_163, DDI.DDI_164, DDI.DDI_165, DDI.DDI_166, DDI.DDI_167, DDI.DDI_168, DDI.DDI_169, DDI.DDI_170, DDI.DDI_171, DDI.DDI_172, DDI.DDI_173, DDI.DDI_174, DDI.DDI_175, DDI.DDI_176};
    public static final DDI[] ACTUAL_RATE_DDIS = new DDI[] {DDI.DDI_2, DDI.DDI_7, DDI.DDI_12, DDI.DDI_17, DDI.DDI_22, DDI.DDI_27, DDI.DDI_32, DDI.DDI_37, DDI.DDI_42, DDI.DDI_47};
    public static final DDI[] SETPOINT_RATE_DDIS = new DDI[] {DDI.DDI_1, DDI.DDI_6, DDI.DDI_11, DDI.DDI_16, DDI.DDI_21, DDI.DDI_26, DDI.DDI_31, DDI.DDI_36, DDI.DDI_41, DDI.DDI_46};
    public static final DDI[] YIELD_DDIS = new DDI[] {DDI.DDI_83, DDI.DDI_84, DDI.DDI_85, DDI.DDI_86, DDI.DDI_87, DDI.DDI_88, DDI.DDI_181, DDI.DDI_182};
    
    /**
     * Calculates a coordinate from a origin with a x and y offset in
     * millimeters and a bearing in decimal degrees from due north ranging from
     * -180 to 180 where 0 is due north. The X-axis is specified as positive in
     * the normal driving direction. The Y-axis is specified as positive to the
     * right side of the device relative to the normal driving direction.
     *
     * @param origin the origin coordinate in WGS84
     * @param xOffset the x offset in millimeters
     * @param yOffset the y offset in millimeters
     * @param bearing bearing in decimal degrees from due north ranging from
     * -180 to 180 where 0 is due north
     * @return
     */
    public static Coordinate getCoordinate(Coordinate origin, double xOffset, double yOffset, double bearing) {
        // pythagoras a2 + b2 = c2, and convert to meters:
        double distance = Math.hypot(xOffset, yOffset) / 1000;
        // angle of the offset, relative to driving direction (positive is to the right):
        double angle = Math.toDegrees(Math.atan2(yOffset, xOffset));
        Point2D calculated = GeometryTools.getDestinationPoint(origin.y, origin.x, normalize(bearing + angle), distance);
        return new Coordinate(calculated.getX(), calculated.getY());
    }

    /**
     * Get the x offset for a specific deviceElement Searches in the provided {@link IsobusTime}
     * for a {@link DataLogValue} which has a reference to the provided
     * DeviceElement and {@link DataLogValue#processDataDDI} that matches {@link DDI#DDI_134}
     *
     * @param isobusTime
     * @param deviceElement
     * @return the found {@link DataLogValue#processDataValue} times the {@link DDI#resolution}
     * or null if not found
     */
    public static Double getXOffsetForDeviceElement(IsobusTime isobusTime, DeviceElement deviceElement) {
        return getProcessDataValueForDeviceElement(isobusTime.getDLV(), deviceElement, DDI.DDI_134);
    }

    /**
     * Get the y offset for a specific deviceElement Searches in the provided {@link IsobusTime}
     * for a {@link DataLogValue} which has a reference to the provided
     * DeviceElement and {@link DataLogValue#processDataDDI} that matches {@link DDI#DDI_134}
     *
     * @param isobusTime
     * @param deviceElement
     * @return the found {@link DataLogValue#processDataValue} times the {@link DDI#resolution}
     * or null if not found
     */
    public static Double getYOffsetForDeviceElement(IsobusTime isobusTime, DeviceElement deviceElement) {
        return getProcessDataValueForDeviceElement(isobusTime.getDLV(), deviceElement, DDI.DDI_135);
    }

    /**
     * Gets the value for the actual work state of an element. possible return
     * values 0 (manual/off) 1 (auto/on) 10 (error) 11 (undefined/not installed)
     *
     * @param isobusTime
     * @param deviceElement
     * @return
     */
    public static Double getActualWorkStateForDeviceElement(List<DataLogValue> dataLogValues, DeviceElement deviceElement) {
        return getProcessDataValueForDeviceElement(dataLogValues, deviceElement, DDI.DDI_141);
    }

    /**
     * Get a {@link DataLogValue#processDataValue} times the {@link DDI#resolution}
     * for a {@link DeviceElement} included in a {@link IsobusTime}
     *
     * @param isobusTime
     * @param deviceElement
     * @param ddi
     * @return the found {@link DataLogValue#processDataValue} times the {@link DDI#resolution}
     * or null if not found
     */
    public static Double getProcessDataValueForDeviceElement(List<DataLogValue> datalogValues, DeviceElement deviceElement, DDI ddi) {
        for (DataLogValue dataLogValue : datalogValues) {
            if (Arrays.equals(ddi.getDDEntityId(), dataLogValue.getProcessDataDDI())) {
                if (deviceElement.equals(dataLogValue.getDeviceElementIdRef())) {
                    return ddi.toImportValue(dataLogValue.getProcessDataValue());
                }
            }
        }
        return null;
    }
    
    /**
     * Find the maximum value present in a timelog for a device element / ddi combination.
     * 
     * @param isobusTimes
     * @param deviceElement
     * @param ddi
     * @return the maximum value or null when no values found.
     */
    public static double getMaxProcessDataValueForDeviceElement(List<IsobusTime> isobusTimes, DeviceElement deviceElement, DDI ddi) {
        Double max = null;
        for (IsobusTime isobusTime : isobusTimes) {
            Double value = getProcessDataValueForDeviceElement(isobusTime.getDLV(), deviceElement, ddi);
            if (value != null && (max == null || value > max)) {
                max = value;
            }
        }
        return max;
    }

    /**
     * Get the application rate for the verbose annotation of application rate.
     * This is when the application amount is annotated on the device element
     * sections them self. Assumes there is only one {@link DeviceElement} of
     * type {@link DeviceElementTypeEnum#DEVICE} in the supplied datalogvalues
     *
     * @param dataLogValues
     * @param ddi
     * @return
     */
    public static Map<DeviceElement, Double> getVerboseWorkStateApplicationRate(Set<DeviceElement> deviceElements, List<DataLogValue> dataLogValues, DDI ddi, boolean distribute) {
        Map<DeviceElement, Double> applicationMap = new HashMap<>();
        //first check if all the sections have there own notation of the provided DDI. if some have and others not we cannot calculate
        Set<DeviceElement> sections = getSectionDeviceElements(deviceElements);
        for (DeviceElement section : sections) {
            //first check if the section has it's own notation of the provided DDI
            Double ownValue = getProcessDataValueForDeviceElement(dataLogValues, section, ddi);
            if (ownValue != null) {
                applicationMap.put(section, ownValue);
            }
        }
        //if all the sections have there own return the map
        if (applicationMap.size() == sections.size()) {
            return applicationMap;
        } else if (applicationMap.isEmpty()) {
            // no own notation for the provided ddi
            //now search for the ddi in the parents
            DeviceElement device = getMainDeviceElement(deviceElements);
            Double totalApplicationAmount = getProcessDataValueForDeviceElement(dataLogValues, device, ddi);
            if (totalApplicationAmount == null) {
                //Not found cannot calculate
                return null;
            }
            //now collect al the section devices which reference the same parent
            //Which have a datalogValue for on of DDI
            Double applicationAmount;
            if (distribute) {
                int onCounter = 0;
                for (DeviceElement section : sections) {
                    Double ddi141Value = getActualWorkStateForDeviceElement(dataLogValues, section);
                    if (ddi141Value != null && ddi141Value == WORK_STATE_ON_AUTO) {
                        onCounter++;
                    }
                }
                applicationAmount = totalApplicationAmount / onCounter;
            } else {
                applicationAmount = totalApplicationAmount;
            }
            
            //Loop again innefficient but necessary
            for (DeviceElement section : sections) {
                Double ddi141Value = getActualWorkStateForDeviceElement(dataLogValues, section);
                Double sectionApplicationAmount = null;
                if (ddi141Value != null && ddi141Value == WORK_STATE_ON_AUTO) {
                    sectionApplicationAmount = applicationAmount;
                }
                applicationMap.put(section, sectionApplicationAmount);
            }
        }
        return applicationMap;
    }

    /**
     * Get the application rate from condensed work state for all {@link DeviceElement}
     * of {@link DeviceElementTypeEnum#SECTION} found in the supplied list of
     * data log values. Will check if one of the {@link DDI#DDI_161} till {@link DDI#DDI_176}
     * are available for the found device element it's parent. If this is the
     * case will check how many nozzles there are and how many are on. Will
     * divide the application rate by that number of found nozzles which are on
     *
     * @param datalogValues
     * @param deviceElement
     * @return the calculated Application rate for the {@link DeviceElement}
     */
    public static Map<DeviceElement, Double> getCondensedWorkStateApplicationRate(Set<DeviceElement> deviceElements, List<DataLogValue> dataLogValues, DDI ddi, boolean distribute) {
        DeviceElement device = getMainDeviceElement(deviceElements);
        if (device != null) {
            //Check if device is on
            Double actualWorkState = getProcessDataValueForDeviceElement(dataLogValues, device, DDI.DDI_141);
            if (actualWorkState != null && WORK_STATE_ON_AUTO == actualWorkState) {
                //get the total application amount for the specified application rate ddi
                Double totalApplicationAmount = getProcessDataValueForDeviceElement(dataLogValues, device, ddi);
                //Only if the total application amount is not null
                if (totalApplicationAmount != null) {
                    Map<DeviceElement, Double> applicationMap = new HashMap<>();
                    //First check if there is a condensed work state at all by checking the first 16
                    Double workState_1_16 = getProcessDataValueForDeviceElement(dataLogValues, device, DDI.DDI_161);
                    if (workState_1_16 != null) {
                        //now collect all the section device elements. This is the max number of condensed work state ddi's there should be
                        //This could be determined by the bit pattern 11 but some suppliers don't implement that correct
                        Set<DeviceElement> sections = getSectionDeviceElements(deviceElements);
                        //Determine in which order the section device elements are based on there y offset DDI_135
                        Map<Double, DeviceElement> yOffsets = new TreeMap<>();
                        for (DeviceElement section : sections) {
                            Double yOffset = getProcessDataValueForDeviceElement(dataLogValues, section, DDI.DDI_135);
                            if (yOffset == null) {
                                //yOffset does not exist cannot calculate
                                return null;
                            }
                            yOffsets.put(yOffset, section);
                        }
                        //Collect the data log values  
                        int numberOfWorkStates = sections.size() / 16 + 1;
                        Double[] values = new Double[numberOfWorkStates];
                        for (int i = 0; i < numberOfWorkStates; i++) {
                            values[i] = getProcessDataValueForDeviceElement(dataLogValues, device, CONDENSED_WORK_STATE_DDIS[i]);
                        }
                        //count how many nozzles or sections are on
                        Double applicationAmount;
                        if (distribute) {
                            int onCounter = 0;
                            for (Double dataLogValue : values) {
                                for (int workState : expandCondensedWorkState(dataLogValue.intValue())) {
                                    if (WORK_STATE_ON_AUTO == workState) {
                                        onCounter++;
                                    }
                                }
                            }
                            applicationAmount = onCounter != 0 ? totalApplicationAmount / onCounter : totalApplicationAmount;
                        } else {
                            applicationAmount = totalApplicationAmount;
                        }
                        
                        Iterator<Double> yOffsetIter = yOffsets.keySet().iterator();
                        //Loop again innefficient but necessary
                        for (Double dataLogValue : values) {
                            for (int workstate : expandCondensedWorkState(dataLogValue.intValue())) {
                                Double sectionApplicationAmount = null;
                                if (WORK_STATE_ON_AUTO == workstate) {
                                    sectionApplicationAmount = applicationAmount;
                                }
                                //Cause of implementatiin fault manufacturer we need to restrict to the amount of DeviceElements available 
                                if (yOffsetIter.hasNext()) {
                                    DeviceElement element = yOffsets.get(yOffsetIter.next());
                                    applicationMap.put(element, sectionApplicationAmount);
                                }
                            }
                        }
                        return applicationMap;
                    }
                }
            }
        }
        return null;
    }

    /**
     * Gets all the device elements of type {@link DeviceElementTypeEnum#SECTION}
     * from a list of {@link DeviceElement}
     *
     * @param deviceElements
     * @return
     */
    public static Set<DeviceElement> getSectionDeviceElements(Set<DeviceElement> deviceElements) {
        Set<DeviceElement> sections = new HashSet<>();
        for (DeviceElement deviceElement : deviceElements) {
            if (DeviceElementTypeEnum.SECTION.equals(deviceElement.getType())) {
                sections.add(deviceElement);
            }
        }
        return sections;
    }

    /**
     * Get the {@link DeviceElement} of type {@link DeviceElementTypeEnum#DEVICE}
     * from the list of {@link DeviceElement}
     *
     * @param deviceElements
     * @return
     */
    public static DeviceElement getMainDeviceElement(Set<DeviceElement> deviceElements) {
        for (DeviceElement deviceElement : deviceElements) {
            if (DeviceElementTypeEnum.DEVICE.equals(deviceElement.getType())) {
                return deviceElement;
            }
        }
        return null;
    }

    /**
     * Tries to find the supplied DDI in the supplied {@link DataLogValue}'s for
     * a {@link DeviceElement} with {@link DeviceElement#id} matching supplied
     * id. If not found will try to find it in one of the parents.
     *
     * @param datalogValues
     * @param id device element object id
     * @param ddi
     * @return value if found else null
     */
    public static Double getProcessDataValueForId(List<DataLogValue> datalogValues, int id, DDI ddi) {
        for (DataLogValue dataLogValue : datalogValues) {
            DeviceElement deviceElement = (DeviceElement) dataLogValue.getDeviceElementIdRef();
            if (deviceElement.getObjectId() == id) {
                Double found = getProcessDataValueForDeviceElement(datalogValues, deviceElement, ddi);
                return found != null ? found : getProcessDataValueForId(datalogValues, deviceElement.getParentObjectId(), ddi);
            }
        }
        return null;
    }

    /**
     * Gets all the device elements which reference the {@link DeviceElement}
     * with {@link DeviceElement#objectId} matching the provided id
     *
     * @param datalogValues
     * @param parentId
     * @return set of found children
     */
    public static Set<DeviceElement> getChildDeviceElements(List<DataLogValue> datalogValues, int id) {
        Set<DeviceElement> children = new HashSet();
        for (DataLogValue dataLogValue : datalogValues) {
            DeviceElement deviceElement = (DeviceElement) dataLogValue.getDeviceElementIdRef();
            if (deviceElement.getParentObjectId() == id) {
                children.add(deviceElement);
            }
        }
        return children;
    }

    /**
     * <p>Method which can read the values of the {@link DDI}'s 161 to 176, and 290 to 305.
     * These 32 bit values are actually 16 2-bit patterns describing the state
     * of a {@link DeviceElement}.</p> the returned list will always have a size
     * of 16
     *
     * @param value
     * @return int array of {@link DeviceState} where the first in the array is
     * the left most device
     */
    public static int[] expandCondensedWorkState(int value) {
        int[] bitPairs = new int[16];
        for (int i = 0; i < 16; i++) {
            bitPairs[i] = value & 3;
            value >>>= 2;
        }
        return bitPairs;
    }
    
    // determine which DDI from a timelog xml header will be shown in rasters
    public static DDI getAppliedRateYieldDDI(Set<DDI> ddis) {
        if (ddis == null || ddis.isEmpty()) { return null; }
        for (DDI ddi : ACTUAL_RATE_DDIS) {
            if (ddis.contains(ddi)) {return ddi; }
        }
        for (DDI ddi : YIELD_DDIS) {
            if (ddis.contains(ddi)) {return ddi; }
        }
        return null;
    }
    
    /**
     * get the setpoint DDI from a set of ddi's
     * @param ddis
     * @return 
     */
    public static DDI getSetpointRateDDI(Set<DDI> ddis){
         if (ddis == null || ddis.isEmpty()) { return null; }
        for (DDI ddi : SETPOINT_RATE_DDIS) {
            if (ddis.contains(ddi)) {return ddi; }
        }
        return null;
    }
            
    
    /**
     * Normalize an angle to the range -180..180.
     * 
     * TODO: to GeometryUtil
     * 
     * @param angle in decimal degrees
     * @return 
     */
    public static double normalize(double angle) {
        while (angle < -180) { angle += 360; }
        while (angle >  180) { angle -= 360; }
        return angle;
    }
    
    /**
     * Find the device properties used in a specific device element.
     * Device process data (DPD) should match (or be a superset of) the DDIs 
     * available in the timelog xml header, device properties (DPT) however are
     * only  available through the taskdata file.
     * 
     * @param taskData
     * @param deviceElement
     * @return list of device properties, possibly empty, not null.
     */
    public static List<DeviceProperty> getDeviceElementProperties(IsobusTaskData taskData, DeviceElement deviceElement) {
        List<DeviceProperty> properties = new ArrayList<>();
        
        if (taskData == null || deviceElement == null) { return properties; }
        
        for (IsobusDevice device : taskData.getDVC()) {
            if (device.getDET().contains(deviceElement)) {
                // compare the properties available on the device with the references used in the element:
                for (DeviceProperty property : device.getDPT()) {
                    for (DeviceObjectReference ref: deviceElement.getDeviceObjectReferences()) {
                        if (property.getId() == ref.getId()) {
                            properties.add(property);
                        }
                    }
                }
            }
        }
        return properties;
    }
    
    /**
     * Find the value of a device property for a specific device element.
     * 
     * @param taskData
     * @param deviceElement
     * @param ddi
     * @return the property value or null if not found.
     */
    public static Double getDevicePropertyValue(IsobusTaskData taskData, DeviceElement deviceElement, DDI ddi) {
        for (DeviceProperty property : getDeviceElementProperties(taskData, deviceElement)) {
            if (Arrays.equals(property.getDDI(), ddi.getDDEntityId())) {
                return ddi.toImportValue(property.getValue());
            }
        }
        return null;
    }

}
